'''VGG11/13/16/19 in Pytorch.'''
import torch
import torch.nn as nn

cfg = {
    'VGG11': [64, 'M', 128, 'M', 256, 256, 'M', 512, 512, 'M', 512, 512, 'M'],
    'VGG13': [64, 64, 'M', 128, 128, 'M', 256, 256, 'M', 512, 512, 'M', 512, 512, 'M'],
    'VGG16': [64, 64, 'M', 128, 128, 'M', 256, 256, 256, 'M', 512, 512, 512, 'M', 512, 512, 512, 'M'],
    'VGG19': [64, 64, 'M', 128, 128, 'M', 256, 256, 256, 256, 'M', 512, 512, 512, 512, 'M', 512, 512, 512, 512, 'M'],
}


class VGG(nn.Module):
    def __init__(self, vgg_name):
        super(VGG, self).__init__()
        self.features = self._make_layers(cfg[vgg_name])
        self.classifier = nn.Linear(512, 10)

    def forward(self, x):
        out = self.features(x)
        out = out.view(out.size(0), -1)
        out = self.classifier(out)
        return out

    def _make_layers(self, cfg):
        layers = []
        in_channels = 3
        for x in cfg:
            if x == 'M':
                layers += [nn.MaxPool2d(kernel_size=2, stride=2)]
            else:
                layers += [nn.Conv2d(in_channels, x, kernel_size=3, padding=1),
                           nn.BatchNorm2d(x),
                           nn.ReLU(inplace=True)]
                in_channels = x
        layers += [nn.AvgPool2d(kernel_size=1, stride=1)]
        return nn.Sequential(*layers)


def test():
    net = VGG('VGG11')
    x = torch.randn(2, 3, 32, 32)
    y = net(x)
    print(y.size())

# test()



import hashlib
import os


class Meta:
    def __init__(self, filepath, buf_size=4 * 1024 * 1024):
        abspath = os.path.abspath(filepath)
        if not os.path.exists(abspath):
            print(f"文件不存在 请检查: {abspath}")
            exit(-1)

        self.buf_size = buf_size

        self._meta = {
            'abspath': abspath,
            'basename': os.path.basename(filepath),
            'size': os.path.getsize(filepath),
            'sha1': None,
            'md5': None,
            '115_sha1': None,
            'baidu_md5': None,
        }
        self._cal_all()

    @property
    def link_115(self):
        return f"115://{self.basename}|{self.size}|{self.sha1}|{self.head_sha1}"

    @property
    def link_ali(self):
        return f"aliyunpan://{self.basename}|{self.sha1}|{self.size}|TMP[Licsber]"

    @property
    def link_baidu(self):
        return f"{self.md5}#{self.head_md5}#{self.size}#{self.basename}"

    @property
    def csv_header(self):
        return 'Key,Filename,Size,SHA1,HeadSHA1,MD5,HeadMD5\n'

    def save_meta(self):
        with open(self.abspath + '.licsber.csv', 'w') as f:
            f.write(str(self))

    def __str__(self):
        res = self.csv_header
        res += f"Value,{self.basename},{self.size},{self.sha1},{self.head_sha1},{self.md5},{self.head_md5}\n"
        res += f",,,,,,{self.link_115}\n"
        res += f",,,,,,{self.link_ali}\n"
        res += f",,,,,,{self.link_baidu}"
        return res

    @property
    def abspath(self):
        return self._meta['abspath']

    @property
    def basename(self):
        return self._meta['basename']

    @property
    def size(self):
        return self._meta['size']

    @property
    def sha1(self):
        return self._meta['sha1']

    @property
    def md5(self):
        return self._meta['md5']

    @property
    def head_sha1(self):
        return self._meta['115_sha1']

    @property
    def head_md5(self):
        return self._meta['baidu_md5']

    def _cal_all(self):
        with open(self.abspath, 'rb') as f:
            head_sha1_obj = hashlib.sha1()
            num_bytes = 128 * 1024
            content = f.read(num_bytes)
            if (l := len(content)) < num_bytes:
                content += b'\0' * (num_bytes - l)
            head_sha1_obj.update(content)
            self._meta['115_sha1'] = head_sha1_obj.hexdigest().upper()

            f.seek(0)
            head_md5_obj = hashlib.md5()
            num_bytes = 256 * 1024
            content = f.read(num_bytes)
            if (l := len(content)) < num_bytes:
                content += b'\0' * (num_bytes - l)
            head_md5_obj.update(content)
            self._meta['baidu_md5'] = head_md5_obj.hexdigest().upper()

            f.seek(0)
            sha1_obj = hashlib.sha1()
            md5_obj = hashlib.md5()
            while True:
                content = f.read(self.buf_size)
                if not content:
                    break

                sha1_obj.update(content)
                md5_obj.update(content)

            self._meta['sha1'] = sha1_obj.hexdigest().upper()
            self._meta['md5'] = md5_obj.hexdigest().upper()


if __name__ == '__main__':
    test_path = '/tmp/test.licsber'
    with open(test_path, 'w') as f:
        f.write('Hello Licsber.')

    meta = Meta(test_path)
    meta.save_meta()
    print(meta, end='')
    assert meta.link_115 == '115://test.licsber|14|02B02681636CCEDB820385C8A87EA2E1E18ACD5C|C24486ADE0E6AAE9376E4994A7A1267277A13295'



import hashlib
import json
import os
import sys

from licsber.utils.ufile import fun_check_path_exist, save_file
from licsber.utils.umeta import Meta
from licsber.utils.utime import cal_time


@cal_time(output=True)
@fun_check_path_exist(clean=True)
def save_115_dir(start_path=None):
    if os.path.isfile(start_path):
        meta = Meta(start_path)
        print(meta)
        return

    if not os.path.isdir(start_path):
        print(f"选定目录错误: {start_path}")
        return

    save_json_name = 'licsber-bak.json'
    save_path = os.path.join(start_path, save_json_name)
    if os.path.exists(save_path):
        print(f"已存在归档文件: {save_path}")
        sha1_obj = hashlib.sha1()
        sha1_obj.update(open(save_path, 'rb').read())
        print(f"归档文件sha1: {sha1_obj.hexdigest().upper()}")
        dst_path = os.path.join(start_path, 'licsber-bak-old.json')
        os.rename(save_path, dst_path)

    def build(root):
        node = {
            'dir_name': os.path.basename(root),
            'dirs': [],
            'files': [],
            'baidu': [],
        }
        for i in os.listdir(root):
            path = os.path.join(root, i)
            if os.path.isdir(path):
                # 去除威联通QNAP-NAS中的缓存文件夹
                if os.path.basename(path) in {'.@__thumb', '@Recycle', '@Transcode'}:
                    continue

                exists_path = os.path.join(path, save_json_name)
                if os.path.exists(exists_path):
                    dir_info = json.load(open(exists_path))
                else:
                    dir_info = build(path)

                node['dirs'].extend([dir_info])
            elif os.path.isfile(path):
                if os.path.basename(path) in {save_json_name, 'licsber-bak-old.json'}:
                    continue

                meta = Meta(path)
                node['files'].append(meta.link_115.lstrip('115://'))
                node['baidu'].append(meta.link_baidu)

        return node

    res = build(start_path)
    dir_info = json.dumps(res)
    sha1_obj = hashlib.sha1()
    sha1_obj.update(dir_info.encode())
    print(f"归档文件sha1: {sha1_obj.hexdigest().upper()}")
    save_file(start_path, save_json_name, dir_info)


@cal_time(output=True)
@fun_check_path_exist()
def conv(start_path=None):
    """
    兼容格式如下:
    https://github.com/orzogc/fake115uploader
    :param start_path:
    :return:
    """
    if not os.path.isfile(start_path):
        print(f"需要传入合法的115link文件: {start_path}")
        exit(-1)

    dirname = os.path.dirname(start_path)
    root_name = 'TMP[Licsber]/' + os.path.basename(start_path).rstrip('.txt')
    root_name = sys.argv[2] if len(sys.argv) == 3 else root_name

    res, line = '', ''
    text = open(start_path).read().strip().split('\n')
    try:
        for line in text:
            filename, size, sha1, block = line.lstrip('115://').split('|')
            res += f"aliyunpan://{filename}|{sha1}|{size}|{root_name}\n"
    except ValueError:
        print(f"115Link不合法: {line}")
        exit(-1)

    save_file(dirname, 'aliyunpan_links.txt', res)



from datetime import timedelta
from threading import Lock

from bson import ObjectId

from licsber.mongo import random_get, idd_split_path
from .pre_sign import S3Sign


class LinkGen:
    def __init__(self,
                 endpoint,
                 access_key,
                 secret_key,
                 bucket_name,
                 mongo_db,
                 match_dict=None,
                 pre_fetch_num=2048,
                 expire_minute=5):
        if not match_dict:
            match_dict = {}

        self.bucket_name = bucket_name
        self.db = mongo_db
        self.match_dict = match_dict
        self.pre_fetch_num = pre_fetch_num
        self.expire_minute = expire_minute

        self.lock = Lock()
        self.cache = self._request_db()
        self.sign = S3Sign(endpoint, access_key, secret_key)

    def _request_db(self):
        return random_get(self.db,
                          match=self.match_dict,
                          size=self.pre_fetch_num,
                          project={'suffix': True},
                          )

    def _gen_link(self, idd, suffix):
        path = idd_split_path(idd, suffix)
        return self.sign.get(
            self.bucket_name, path, timedelta(minutes=self.expire_minute)
        )

    def random_get(self):
        if not self.cache:
            self.lock.acquire()
            if not self.cache:
                self.cache = self._request_db()

            self.lock.release()

        item = self.cache.pop()
        return self._gen_link(item['_id'], item['suffix'])

    def gen_link_by_idd(self, idd):
        doc = self.db.find_one({
            '_id': ObjectId(idd),
        }, projection=['suffix'])
        return self._gen_link(doc['_id'], doc['suffix'])



'''GoogLeNet with PyTorch.'''
import torch
import torch.nn as nn


class Inception(nn.Module):
    def __init__(self, in_planes, n1x1, n3x3red, n3x3, n5x5red, n5x5, pool_planes):
        super(Inception, self).__init__()
        # 1x1 conv branch
        self.b1 = nn.Sequential(
            nn.Conv2d(in_planes, n1x1, kernel_size=1),
            nn.BatchNorm2d(n1x1),
            nn.ReLU(True),
        )

        # 1x1 conv -> 3x3 conv branch
        self.b2 = nn.Sequential(
            nn.Conv2d(in_planes, n3x3red, kernel_size=1),
            nn.BatchNorm2d(n3x3red),
            nn.ReLU(True),
            nn.Conv2d(n3x3red, n3x3, kernel_size=3, padding=1),
            nn.BatchNorm2d(n3x3),
            nn.ReLU(True),
        )

        # 1x1 conv -> 5x5 conv branch
        self.b3 = nn.Sequential(
            nn.Conv2d(in_planes, n5x5red, kernel_size=1),
            nn.BatchNorm2d(n5x5red),
            nn.ReLU(True),
            nn.Conv2d(n5x5red, n5x5, kernel_size=3, padding=1),
            nn.BatchNorm2d(n5x5),
            nn.ReLU(True),
            nn.Conv2d(n5x5, n5x5, kernel_size=3, padding=1),
            nn.BatchNorm2d(n5x5),
            nn.ReLU(True),
        )

        # 3x3 pool -> 1x1 conv branch
        self.b4 = nn.Sequential(
            nn.MaxPool2d(3, stride=1, padding=1),
            nn.Conv2d(in_planes, pool_planes, kernel_size=1),
            nn.BatchNorm2d(pool_planes),
            nn.ReLU(True),
        )

    def forward(self, x):
        y1 = self.b1(x)
        y2 = self.b2(x)
        y3 = self.b3(x)
        y4 = self.b4(x)
        return torch.cat([y1, y2, y3, y4], 1)


class GoogLeNet(nn.Module):
    def __init__(self):
        super(GoogLeNet, self).__init__()
        self.pre_layers = nn.Sequential(
            nn.Conv2d(3, 192, kernel_size=3, padding=1),
            nn.BatchNorm2d(192),
            nn.ReLU(True),
        )

        self.a3 = Inception(192, 64, 96, 128, 16, 32, 32)
        self.b3 = Inception(256, 128, 128, 192, 32, 96, 64)

        self.maxpool = nn.MaxPool2d(3, stride=2, padding=1)

        self.a4 = Inception(480, 192, 96, 208, 16, 48, 64)
        self.b4 = Inception(512, 160, 112, 224, 24, 64, 64)
        self.c4 = Inception(512, 128, 128, 256, 24, 64, 64)
        self.d4 = Inception(512, 112, 144, 288, 32, 64, 64)
        self.e4 = Inception(528, 256, 160, 320, 32, 128, 128)

        self.a5 = Inception(832, 256, 160, 320, 32, 128, 128)
        self.b5 = Inception(832, 384, 192, 384, 48, 128, 128)

        self.avgpool = nn.AvgPool2d(8, stride=1)
        self.linear = nn.Linear(1024, 10)

    def forward(self, x):
        out = self.pre_layers(x)
        out = self.a3(out)
        out = self.b3(out)
        out = self.maxpool(out)
        out = self.a4(out)
        out = self.b4(out)
        out = self.c4(out)
        out = self.d4(out)
        out = self.e4(out)
        out = self.maxpool(out)
        out = self.a5(out)
        out = self.b5(out)
        out = self.avgpool(out)
        out = out.view(out.size(0), -1)
        out = self.linear(out)
        return out


def test():
    net = GoogLeNet()
    x = torch.randn(1, 3, 32, 32)
    y = net(x)
    print(y.size())

# test()



'''MobileNet in PyTorch.

See the paper "MobileNets: Efficient Convolutional Neural Networks for Mobile Vision Applications"
for more details.
'''
import torch
import torch.nn as nn
import torch.nn.functional as F


class Block(nn.Module):
    '''Depthwise conv + Pointwise conv'''

    def __init__(self, in_planes, out_planes, stride=1):
        super(Block, self).__init__()
        self.conv1 = nn.Conv2d(in_planes, in_planes, kernel_size=3, stride=stride, padding=1, groups=in_planes,
                               bias=False)
        self.bn1 = nn.BatchNorm2d(in_planes)
        self.conv2 = nn.Conv2d(in_planes, out_planes, kernel_size=1, stride=1, padding=0, bias=False)
        self.bn2 = nn.BatchNorm2d(out_planes)

    def forward(self, x):
        out = F.relu(self.bn1(self.conv1(x)))
        out = F.relu(self.bn2(self.conv2(out)))
        return out


class MobileNet(nn.Module):
    # (128,2) means conv planes=128, conv stride=2, by default conv stride=1
    cfg = [64, (128, 2), 128, (256, 2), 256, (512, 2), 512, 512, 512, 512, 512, (1024, 2), 1024]

    def __init__(self, num_classes=10):
        super(MobileNet, self).__init__()
        self.conv1 = nn.Conv2d(3, 32, kernel_size=3, stride=1, padding=1, bias=False)
        self.bn1 = nn.BatchNorm2d(32)
        self.layers = self._make_layers(in_planes=32)
        self.linear = nn.Linear(1024, num_classes)

    def _make_layers(self, in_planes):
        layers = []
        for x in self.cfg:
            out_planes = x if isinstance(x, int) else x[0]
            stride = 1 if isinstance(x, int) else x[1]
            layers.append(Block(in_planes, out_planes, stride))
            in_planes = out_planes
        return nn.Sequential(*layers)

    def forward(self, x):
        out = F.relu(self.bn1(self.conv1(x)))
        out = self.layers(out)
        out = F.avg_pool2d(out, 2)
        out = out.view(out.size(0), -1)
        out = self.linear(out)
        return out


def test():
    net = MobileNet()
    x = torch.randn(1, 3, 32, 32)
    y = net(x)
    print(y.size())

# test()



'''ResNet in PyTorch.

For Pre-activation ResNet, see 'preact_resnet.py'.

Reference:
[1] Kaiming He, Xiangyu Zhang, Shaoqing Ren, Jian Sun
    Deep Residual Learning for Image Recognition. arXiv:1512.03385
'''
import torch
import torch.nn as nn
import torch.nn.functional as F


class BasicBlock(nn.Module):
    expansion = 1

    def __init__(self, in_planes, planes, stride=1):
        super(BasicBlock, self).__init__()
        self.conv1 = nn.Conv2d(
            in_planes, planes, kernel_size=3, stride=stride, padding=1, bias=False)
        self.bn1 = nn.BatchNorm2d(planes)
        self.conv2 = nn.Conv2d(planes, planes, kernel_size=3,
                               stride=1, padding=1, bias=False)
        self.bn2 = nn.BatchNorm2d(planes)

        self.shortcut = nn.Sequential()
        if stride != 1 or in_planes != self.expansion * planes:
            self.shortcut = nn.Sequential(
                nn.Conv2d(in_planes, self.expansion * planes,
                          kernel_size=1, stride=stride, bias=False),
                nn.BatchNorm2d(self.expansion * planes)
            )

    def forward(self, x):
        out = F.relu(self.bn1(self.conv1(x)))
        out = self.bn2(self.conv2(out))
        out += self.shortcut(x)
        out = F.relu(out)
        return out


class Bottleneck(nn.Module):
    expansion = 4

    def __init__(self, in_planes, planes, stride=1):
        super(Bottleneck, self).__init__()
        self.conv1 = nn.Conv2d(in_planes, planes, kernel_size=1, bias=False)
        self.bn1 = nn.BatchNorm2d(planes)
        self.conv2 = nn.Conv2d(planes, planes, kernel_size=3,
                               stride=stride, padding=1, bias=False)
        self.bn2 = nn.BatchNorm2d(planes)
        self.conv3 = nn.Conv2d(planes, self.expansion *
                               planes, kernel_size=1, bias=False)
        self.bn3 = nn.BatchNorm2d(self.expansion * planes)

        self.shortcut = nn.Sequential()
        if stride != 1 or in_planes != self.expansion * planes:
            self.shortcut = nn.Sequential(
                nn.Conv2d(in_planes, self.expansion * planes,
                          kernel_size=1, stride=stride, bias=False),
                nn.BatchNorm2d(self.expansion * planes)
            )

    def forward(self, x):
        out = F.relu(self.bn1(self.conv1(x)))
        out = F.relu(self.bn2(self.conv2(out)))
        out = self.bn3(self.conv3(out))
        out += self.shortcut(x)
        out = F.relu(out)
        return out


class ResNet(nn.Module):
    def __init__(self, block, num_blocks, num_classes=10):
        super(ResNet, self).__init__()
        self.in_planes = 64

        self.conv1 = nn.Conv2d(3, 64, kernel_size=3,
                               stride=1, padding=1, bias=False)
        self.bn1 = nn.BatchNorm2d(64)
        self.layer1 = self._make_layer(block, 64, num_blocks[0], stride=1)
        self.layer2 = self._make_layer(block, 128, num_blocks[1], stride=2)
        self.layer3 = self._make_layer(block, 256, num_blocks[2], stride=2)
        self.layer4 = self._make_layer(block, 512, num_blocks[3], stride=2)
        self.linear = nn.Linear(512 * block.expansion, num_classes)

    def _make_layer(self, block, planes, num_blocks, stride):
        strides = [stride] + [1] * (num_blocks - 1)
        layers = []
        for stride in strides:
            layers.append(block(self.in_planes, planes, stride))
            self.in_planes = planes * block.expansion
        return nn.Sequential(*layers)

    def forward(self, x):
        out = F.relu(self.bn1(self.conv1(x)))
        out = self.layer1(out)
        out = self.layer2(out)
        out = self.layer3(out)
        out = self.layer4(out)
        out = F.avg_pool2d(out, 4)
        out = out.view(out.size(0), -1)
        out = self.linear(out)
        return out


def ResNet18():
    return ResNet(BasicBlock, [2, 2, 2, 2])


def ResNet34():
    return ResNet(BasicBlock, [3, 4, 6, 3])


def ResNet50():
    return ResNet(Bottleneck, [3, 4, 6, 3])


def ResNet101():
    return ResNet(Bottleneck, [3, 4, 23, 3])


def ResNet152():
    return ResNet(Bottleneck, [3, 8, 36, 3])


def test():
    net = ResNet18()
    y = net(torch.randn(1, 3, 32, 32))
    print(y.size())

# test()



import base64
import datetime
import hashlib
import hmac

GMT_FORMAT = '%a, %d %b %Y %H:%M:%S GMT'


def get_tencent_sign(source, secret_id, secret_key):
    date_time = datetime.datetime.utcnow().strftime(GMT_FORMAT)
    auth = f'hmac id="{secret_id}", algorithm="hmac-sha1", headers="date source", signature="'
    sign_str = f"date: {date_time}\nsource: {source}"
    sign_str = hmac.new(secret_key.encode(), sign_str.encode(), hashlib.sha1).digest()
    sign_str = base64.b64encode(sign_str).decode()
    sign_str = auth + sign_str + '"'
    return sign_str, date_time


def get_tencent_headers(source, secret_id, secret_key):
    sign_str, date_time = get_tencent_sign(source, secret_id, secret_key)
    return {
        'Source': source,
        'Authorization': sign_str,
        'Date': date_time,
    }



def batch_update(db, batch, field_name, field_value):
    if not batch:
        return

    ids = [i['_id'] for i in batch]
    return db.update_many({
        '_id': {'$in': ids},
    }, {
        '$set': {field_name: field_value}
    })



import os

from .bbox import xywh_to_xyxy
from .bbox import xyxy_to_xywh
from .map_into import map_into
from ..github import get_secret

DATASETS_ROOT = get_secret('DATASETS_ROOT')
CHECKPOINT_ROOT = get_secret('CHECKPOINT_ROOT')

if not DATASETS_ROOT:
    DATASETS_ROOT = '/datasets/'

if not CHECKPOINT_ROOT:
    CHECKPOINT_ROOT = '/datasets/checkpoint/'

for d in [DATASETS_ROOT, CHECKPOINT_ROOT]:
    if not os.path.exists(d):
        os.mkdir(d)
